from .net_helper import is_str

INSERT_DATASET = """
    INSERT INTO DataSets(DataSetName, NodeId, {columns})
    VALUES (:name, :NodeId, {values})
"""
DELETE_DATASET = """
    DELETE FROM DataSets WHERE DataSetName = :name
"""

SELECT_ALL_DATASET_NAMES = """
    SELECT DISTINCT DataSetName,NodeId FROM DataSets
"""

SELECT_DATASET_BY_NAME = """
    SELECT * FROM DataSets WHERE DataSetName = :name
"""

# node queries
# sql_queries.py

INSERT_NODE_QUERY = (
    "INSERT INTO Nodes(NodeName,NetId,train) VALUES ('{}',{}, {})"
)
DELETE_NODE_QUERY = "DELETE FROM Nodes WHERE NodeName = '{}'"
SELECT_MASTER_COLUMNS_QUERY = "SELECT * FROM MasterDataset LIMIT 1"
SELECT_DATASET_BY_COLUMN_QUERY = "SELECT * FROM MasterDataset WHERE {} = '{}'"
SELECT_DATASET_BY_NODE_ID_QUERY = "SELECT * FROM DataSets WHERE NodeId = {}"

SELECT_ALL_DATASETS_QUERY = "SELECT DISTINCT DataSetName,NodeName FROM DataSets,Nodes WHERE Nodes.NodeName = '{}' and Nodes.NodeId = DataSets.NodeId"
SELECT_ALL_NODES_QUERY = "SELECT * FROM Nodes"


# SQL query to insert a new network
INSERT_NETWORK_QUERY = "INSERT INTO Networks(NetName) VALUES (:name)"

# SQL query to delete a network
DELETE_NETWORK_QUERY = "DELETE FROM Networks WHERE NetName = '{name}'"

# SQL query to delete a network
GET_NETWORK_QUERY = "SELECT * FROM Networks WHERE NetName = '{name}'"


# SQL query to update a network
UPDATE_NETWORK_QUERY = (
    "UPDATE Networks SET FLsetupId = {FLsetupId} WHERE NetId = {id}"
)

# SQL query to retrieve all nodes for a network
LIST_ALL_NODES_QUERY = """
SELECT Nodes.NodeName, Networks.NetName 
FROM Nodes 
JOIN Networks ON Networks.NetId = Nodes.NetId 
WHERE Networks.NetName = :name
"""

# SQL query to create the MasterDataset table (SQLite-compatible)
CREATE_MASTER_DATASET_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS MasterDataset (
    PatientId INTEGER PRIMARY KEY AUTOINCREMENT,
    {}
);
"""


# SQL query to create the datasets table (SQLite-compatible)
CREATE_DATASETS_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS Datasets (
    DataSetId INTEGER PRIMARY KEY AUTOINCREMENT, 
    DataSetName VARCHAR(255), 
    NodeId INT,
    {}
);
"""



# SQL query to insert dataset values
INSERT_DATASET_VALUES_QUERY = "INSERT INTO MasterDataset({columns}, NodeId) VALUES ('{name}', {nodeId}, {values})"


# FL setup_queries
# sql_queries.py

CREATE_FLSETUP_QUERY = """
    INSERT INTO FLsetup (name, description, creation_date, NetId, column_name)
    VALUES (:name, :description, :creation_date, :net_id, :column_name)
"""

DELETE_FLSETUP_QUERY = """
    DELETE FROM FLsetup
    WHERE name = :name
"""

UPDATE_FLSETUP_QUERY = UPDATE_NETWORK_QUERY = (
    "UPDATE FLsetup SET column_name ='{column_name}' WHERE name ='{FLsetupName}'"
)


READ_SETUP_QUERY = """
    SELECT * FROM FLsetup
    WHERE FLsetupId = :flsetup_id
"""

READ_ALL_SETUPS_QUERY = """
    SELECT * FROM FLsetup
"""

READ_NETWORK_BY_ID_QUERY = """
    SELECT * FROM Networks
    WHERE NetId = :net_id
"""

READ_DISTINCT_NODES_QUERY = """
SELECT DISTINCT {} FROM MasterDataset 
"""


# FederatedDataset Queries
INSERT_FLDATASET_QUERY = (
    "INSERT INTO FedDatasets(name, FLsetupId) VALUES (:name, :FLsetupId)"
)
DELETE_FLDATASET_BY_SETUP_AND_PIPELINE_QUERY = "DELETE FROM FedDatasets WHERE FLsetupId = :FLsetupId AND FLpipeId = :FLpipeId"


UPDATE_FLDATASET_QUERY = (
    "UPDATE FedDatasets SET FLpipeId = :FLpipeId WHERE FedId = :FedId"
)
SELECT_FLDATASET_BY_NAME_QUERY = "SELECT * FROM FedDatasets WHERE name = :name"

CREATE_FLPIPELINE_QUERY = """
INSERT INTO FLpipeline (name, description, creation_date, results)
VALUES ('{name}', '{description}', '{creation_date}', '{result}')
"""
DELETE_FLPIPELINE_QUERY = "DELETE FROM FLpipeline WHERE name = '{name}'"

SELECT_FLPIPELINE_QUERY = "SELECT FROM FLpipeline WHERE name = '{name}'"

CREATE_TEST_RESULTS_QUERY = """
INSERT INTO testResults (pipelineid, nodename, confusionmatrix, accuracy , sensivity, ppv , npv , f1score , fpr , tpr )
VALUES ('{pipelineId}', '{nodeName}', '{confusion_matrix}', '{accuracy}' , '{sensivity}' , '{ppv}' , '{npv}' , '{f1score}' , '{fpr}' , '{tpr}')
"""



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from .utils import *

# Replace this with your actual code for data collection
results_dict = {
    ("LR: 0.001, Optimizer: Adam", "accuracy"): [0.85, 0.89, 0.92, 0.94, ...],
    ("LR: 0.001, Optimizer: Adam", "loss"): [0.2, 0.15, 0.1, 0.08, ...],
    ("LR: 0.01, Optimizer: SGD", "accuracy"): [0.88, 0.91, 0.93, 0.95, ...],
    ("LR: 0.01, Optimizer: SGD", "loss"): [0.18, 0.13, 0.09, 0.07, ...],
    ("LR: 0.1, Optimizer: Adam", "accuracy"): [0.82, 0.87, 0.91, 0.93, ...],
    ("LR: 0.1, Optimizer: Adam", "loss"): [0.25, 0.2, 0.15, 0.12, ...],
}
"""
server should have:
 #len = num of rounds
  self.accuracies
  self.losses
  
Client should have
  # len = num of epochs
  self.accuracies
  self.losses
  self.epsilons
  self.deltas
  
#common things : LR,SGD, Aggregation
  
"""


class AccuracyLossPlotter:
    """
    A utility class for plotting accuracy and loss metrics based on experiment results.

    Args:
        results_dict (dict): Dictionary containing experiment results organized by parameters and metrics.

    Attributes:
        results_dict (dict): Dictionary containing experiment results organized by parameters and metrics.
        parameters (list): List of unique parameters in the experiment results.
        metrics (list): List of unique metrics in the experiment results.
        iterations (range): Range of iterations (rounds or epochs) in the experiment.
    """

    def __init__(self, results_dict):
        """
        Initialize the AccuracyLossPlotter with experiment results.

        Args:
            results_dict (dict): Dictionary containing experiment results organized by parameters and metrics.
        """
        self.results_dict = results_dict
        self.parameters = list(
            set([param[0] for param in results_dict.keys()])
        )
        self.metrics = list(set([param[1] for param in results_dict.keys()]))
        self.iterations = range(1, len(list(results_dict.values())[0]) + 1)

    def plot_accuracy_loss(self):
        """
        Plot accuracy and loss metrics for different parameters.
        """

        plt.figure(figsize=(8, 6))

        for param in self.parameters:
            for metric in self.metrics:
                key = (param, metric)
                values = self.results_dict[key]
                plt.plot(
                    self.iterations,
                    values,
                    label=f"{param} ({metric})",
                    marker="o",
                    linestyle="-",
                )

            plt.xlabel("Rounds")
            plt.ylabel("Accuracy / Loss")
            plt.title("Accuracy and Loss by Parameters")
            plt.legend()
            plt.grid(True)
            plt.show()

    @staticmethod
    def plot_global_confusion_matrix(pipeline_name: str):
        """
        Plot a global confusion matrix based on pipeline results.

        Args:
            pipeline_name (str): Name of the pipeline.

        Returns:
            None
        """
        # Get the id of the pipeline by name
        pipeline_id = get_pipeline_from_name(pipeline_name)
        # get the confusion matrix pf the pipeline
        confusion_matrix = get_pipeline_confusion_matrix(pipeline_id)

        # Extracting confusion matrix values
        TP = confusion_matrix['TP']
        FP = confusion_matrix['FP']
        FN = confusion_matrix['FN']
        TN = confusion_matrix['TN']

        # Creating a matrix for visualization
        matrix = [[TN, FP],
                  [FN, TP]]

        # Plotting the confusion matrix as a heatmap
        plt.figure(figsize=(6, 4))
        sns.heatmap(matrix, annot=True, fmt='d', cmap='Blues',
                    xticklabels=['Predicted Negative', 'Predicted Positive'],
                    yticklabels=['Actual Negative', 'Actual Positive'])
        plt.title('Global Confusion Matrix')
        plt.xlabel('Predicted label')
        plt.ylabel('True label')
        plt.tight_layout()

        # Display the confusion matrix heatmap
        plt.show()

    @staticmethod
    def plot_confusion_Matrix_by_node(node_name: str, pipeline_name: str):
        """
        Plot a confusion matrix for a specific node in the pipeline.

        Args:
            node_name (str): Name of the node.
            pipeline_name (str): Name of the pipeline.

        Returns:
            None
        """

        # Get the id of the pipeline by name
        pipeline_id = get_pipeline_from_name(pipeline_name)
        # get the confusion matrix pf the pipeline
        confusion_matrix = get_node_confusion_matrix(
            pipeline_id, node_name=node_name)

        # Extracting confusion matrix values
        TP = confusion_matrix['TP']
        FP = confusion_matrix['FP']
        FN = confusion_matrix['FN']
        TN = confusion_matrix['TN']

        # Creating a matrix for visualization
        matrix = [[TN, FP],
                  [FN, TP]]

        # Plotting the confusion matrix as a heatmap
        plt.figure(figsize=(6, 4))
        sns.heatmap(matrix, annot=True, fmt='d', cmap='Blues',
                    xticklabels=['Predicted Negative', 'Predicted Positive'],
                    yticklabels=['Actual Negative', 'Actual Positive'])
        plt.title('Confusion Matrix of node: '+node_name)
        plt.xlabel('Predicted label')
        plt.ylabel('True label')
        plt.tight_layout()

        # Display the confusion matrix heatmap
        plt.show()
        return
    
    @staticmethod
    def plot_classification_report(pipeline_name: str):
        """
        Plot a comparison of classification report metrics between nodes.

        Args:
            pipeline_name (str): Name of the pipeline.

        Returns:
            None
        """

        colors = ['#FF5733', '#6A5ACD', '#3CB371', '#FFD700', '#FFA500', '#8A2BE2', '#00FFFF', '#FF00FF', '#A52A2A', '#00FF00']

        # Get the id of the pipeline by name
        pipeline_id = get_pipeline_from_name(pipeline_name)

        pipeline_results = get_pipeline_result(pipeline_id)

        nodesList = pipeline_results['nodename']
        classificationReports = []

        for index, node in enumerate(nodesList):
            classificationReports.append({
                'Accuracy': pipeline_results['accuracy'][index],
                'Sensitivity/Recall': pipeline_results['sensivity'][index],
                'PPV/Precision': pipeline_results['ppv'][index],
                'NPV': pipeline_results['npv'][index],
                'F1-score': pipeline_results['f1score'][index],
                'False positive rate': pipeline_results['fpr'][index],
                'True positive rate': pipeline_results['tpr'][index]
            })

        metric_labels = list(classificationReports[0].keys())  # Assuming both reports have the same keys

        # Set the positions of the bars on the x-axis
        x = np.arange(len(metric_labels))

        # Set the width of the bars
        width = 0.35

        plt.figure(figsize=(12, 6))

        for index, report in enumerate(classificationReports):
            metric = list(report.values())
            plt.bar(x + (index - len(nodesList) / 2) * width / len(nodesList), metric, width / len(nodesList),
                    label=nodesList[index], color=colors[index % len(colors)])

        # Adding labels, title, and legend
        plt.xlabel('Metrics')
        plt.ylabel('Values')
        plt.title('Comparison of Classification Report Metrics between Nodes')
        plt.xticks(ticks=x, labels=metric_labels, rotation=45)
        plt.legend()

        # Show plot
        plt.tight_layout()
        plt.show()

        return 



#!/usr/bin/env python3

import pkg_resources
import torch
import yaml
from sklearn.metrics import *
from yaml.loader import SafeLoader


from MEDfl.NetManager.database_connector import DatabaseManager

# from scripts.base import *
import json


import pandas as pd
import numpy as np

import os
import configparser

import subprocess
import ast

from sqlalchemy import text


# Get the directory of the current script
current_directory = os.path.dirname(os.path.abspath(__file__))

# Load configuration from the config file
yaml_path = os.path.join(current_directory, 'params.yaml')

with open(yaml_path) as g:
    params = yaml.load(g, Loader=SafeLoader)

# global_yaml_path = pkg_resources.resource_filename(__name__, "../../global_params.yaml")
# with open(global_yaml_path) as g:
#     global_params = yaml.load(g, Loader=SafeLoader)


# Default path for the config file
DEFAULT_CONFIG_PATH = 'db_config.ini'


def load_db_config_dep():
    config = os.environ.get('MEDfl_DB_CONFIG')

    if config:
        return ast.literal_eval(config)
    else:
        raise ValueError(f"MEDfl db config not found")

# Function to allow users to set config path programmatically


def set_db_config_dep(config_path):
    config = configparser.ConfigParser()
    config.read(config_path)
    if (config['sqllite']):
        os.environ['MEDfl_DB_CONFIG'] = str(dict(config['sqllite']))
    else:
        raise ValueError(f"mysql key not found in file '{config_path}'")



def load_db_config():
    """Read a dictionary from an environment variable."""
    obj_str = os.getenv("MEDfl_DB_CONFIG")
    if obj_str is not None:
        return ast.literal_eval(obj_str)
    else:
        raise ValueError(f"Environment variable MEDfl_DB_CONFIG not found")

# Function to allow users to set config path programmatically


def set_db_config(config_path):
    obj = {"database" : config_path}

    """Store a dictionary as a string in an environment variable."""
    obj_str = str(obj)
    os.environ['MEDfl_DB_CONFIG'] = obj_str

    




# Create databas


def create_MEDfl_db():
    script_path = os.path.join(os.path.dirname(
        __file__), 'scripts', 'create_db.sh')
    subprocess.run(['sh', script_path], check=True)


def custom_classification_report(y_true, y_pred_prob):
    """
    Compute custom classification report metrics including accuracy, sensitivity, specificity, precision, NPV,
    F1-score, false positive rate, and true positive rate.

    Args:
        y_true (array-like): True labels.
        y_pred (array-like): Predicted labels.

    Returns:
        dict: A dictionary containing custom classification report metrics.
    """
    y_pred = (y_pred_prob).round(
    )  # Round absolute values of predicted probabilities to the nearest integer

    auc = roc_auc_score(y_true, y_pred_prob)  # Calculate AUC

    tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()

    # Accuracy
    denominator_acc = tp + tn + fp + fn
    acc = (tp + tn) / denominator_acc if denominator_acc != 0 else 0.0

    # Sensitivity/Recall
    denominator_sen = tp + fn
    sen = tp / denominator_sen if denominator_sen != 0 else 0.0

    # Specificity
    denominator_sp = tn + fp
    sp = tn / denominator_sp if denominator_sp != 0 else 0.0

    # PPV/Precision
    denominator_ppv = tp + fp
    ppv = tp / denominator_ppv if denominator_ppv != 0 else 0.0

#     NPV
    denominator_npv = tn + fn
    npv = tn / denominator_npv if denominator_npv != 0 else 0.0

    # F1 Score
    denominator_f1 = sen + ppv
    f1 = 2 * (sen * ppv) / denominator_f1 if denominator_f1 != 0 else 0.0

    # False Positive Rate
    denominator_fpr = fp + tn
    fpr = fp / denominator_fpr if denominator_fpr != 0 else 0.0

    # True Positive Rate
    denominator_tpr = tp + fn
    tpr = tp / denominator_tpr if denominator_tpr != 0 else 0.0

    return {
        "confusion matrix": {"TP": tp, "FP": fp, "FN": fn, "TN": tn},
        "Accuracy": round(acc, 3),
        "Sensitivity/Recall": round(sen, 3),
        "Specificity": round(sp, 3),
        "PPV/Precision": round(ppv, 3),
        "NPV": round(npv, 3),
        "F1-score": round(f1, 3),
        "False positive rate": round(fpr, 3),
        "True positive rate": round(tpr, 3),
        "auc": auc
    }


def test(model, test_loader, device=torch.device("cpu")):
    """
    Evaluate a model using a test loader and return a custom classification report.

    Args:
        model (torch.nn.Module): PyTorch model to evaluate.
        test_loader (torch.utils.data.DataLoader): DataLoader for the test dataset.
        device (torch.device, optional): Device for model evaluation. Default is "cpu".

    Returns:
        dict: A dictionary containing custom classification report metrics.
    """

    model.eval()
    with torch.no_grad():
        X_test, y_test = test_loader.dataset[:][0].to(
            device), test_loader.dataset[:][1].to(device)
        y_hat_prob = torch.squeeze(model(X_test), 1).cpu()

    return custom_classification_report(y_test.cpu().numpy(), y_hat_prob.cpu().numpy())


column_map = {"object": "VARCHAR(255)", "int64": "INT", "float64": "FLOAT"}


def empty_db():
    """
    Empty the database by deleting records from multiple tables and resetting auto-increment counters.

    Returns:
        None
    """
    db_manager = DatabaseManager()
    db_manager.connect()
    my_eng = db_manager.get_connection()

    # my_eng.execute(text(f"DELETE FROM  {'DataSets'}"))
    my_eng.execute(text(f"DELETE FROM {'Nodes'}"))
    my_eng.execute(text(f"DELETE FROM {'FedDatasets'}"))
    my_eng.execute(text(f"DELETE FROM {'Networks'}"))
    my_eng.execute(text(f"DELETE FROM {'FLsetup'}"))

    my_eng.execute(text(f"DELETE FROM {'FLpipeline'}"))
    my_eng.execute(text(f"ALTER TABLE {'Nodes'} AUTO_INCREMENT = 1"))
    my_eng.execute(text(f"ALTER TABLE {'Networks'} AUTO_INCREMENT = 1"))
    my_eng.execute(text(f"ALTER TABLE {'FedDatasets'} AUTO_INCREMENT = 1"))
    my_eng.execute(text(f"ALTER TABLE {'FLsetup'} AUTO_INCREMENT = 1"))
    my_eng.execute(text(f"ALTER TABLE {'FLpipeline'} AUTO_INCREMENT = 1"))
    my_eng.execute(text(f"DELETE FROM {'testResults'}"))
    my_eng.execute(text(f"DROP TABLE IF EXISTS {'MasterDataset'}"))
    my_eng.execute(text(f"DROP TABLE IF EXISTS {'DataSets'}"))


def get_pipeline_from_name(name):
    """
    Get the pipeline ID from its name in the database.

    Args:
        name (str): Name of the pipeline.

    Returns:
        int: ID of the pipeline.
    """
    db_manager = DatabaseManager()
    db_manager.connect()
    my_eng = db_manager.get_connection()

    NodeId = int(
        pd.read_sql(
            text(f"SELECT id FROM FLpipeline WHERE name = '{name}'"), my_eng
        ).iloc[0, 0]
    )
    return NodeId


def get_pipeline_confusion_matrix(pipeline_id):
    """
    Get the global confusion matrix for a pipeline based on test results.

    Args:
        pipeline_id (int): ID of the pipeline.

    Returns:
        dict: A dictionary representing the global confusion matrix.
    """
    db_manager = DatabaseManager()
    db_manager.connect()
    my_eng = db_manager.get_connection()

    data = pd.read_sql(
        text(
            f"SELECT confusionmatrix FROM testResults WHERE pipelineid = '{pipeline_id}'"), my_eng
    )

    # Convert the column of strings into a list of dictionaries representing confusion matrices
    confusion_matrices = [
        json.loads(matrix.replace("'", "\"")) for matrix in data['confusionmatrix']
    ]

    # Initialize variables for global confusion matrix
    global_TP = global_FP = global_FN = global_TN = 0

    # Iterate through each dictionary and sum the corresponding values for each category
    for matrix in confusion_matrices:
        global_TP += matrix['TP']
        global_FP += matrix['FP']
        global_FN += matrix['FN']
        global_TN += matrix['TN']

    # Create a global confusion matrix as a dictionary
    global_confusion_matrix = {
        'TP': global_TP,
        'FP': global_FP,
        'FN': global_FN,
        'TN': global_TN
    }
    # Return the list of dictionaries representing confusion matrices
    return global_confusion_matrix


def get_node_confusion_matrix(pipeline_id, node_name):
    """
    Get the confusion matrix for a specific node in a pipeline based on test results.

    Args:
        pipeline_id (int): ID of the pipeline.
        node_name (str): Name of the node.

    Returns:
        dict: A dictionary representing the confusion matrix for the specified node.
    """
    db_manager = DatabaseManager()
    db_manager.connect()
    my_eng = db_manager.get_connection()

    data = pd.read_sql(
        text(
            f"SELECT confusionmatrix FROM testResults WHERE pipelineid = '{pipeline_id}' AND nodename = '{node_name}'"), my_eng
    )

    # Convert the column of strings into a list of dictionaries representing confusion matrices
    confusion_matrices = [
        json.loads(matrix.replace("'", "\"")) for matrix in data['confusionmatrix']
    ]

    # Return the list of dictionaries representing confusion matrices
    return confusion_matrices[0]


def get_pipeline_result(pipeline_id):
    """
    Get the test results for a pipeline.

    Args:
        pipeline_id (int): ID of the pipeline.

    Returns:
        pandas.DataFrame: DataFrame containing test results for the specified pipeline.
    """
    db_manager = DatabaseManager()
    db_manager.connect()
    my_eng = db_manager.get_connection()

    data = pd.read_sql(
        text(
            f"SELECT * FROM testResults WHERE pipelineid = '{pipeline_id}'"), my_eng
    )
    return data



import unittest
from datetime import datetime
from unittest.mock import Mock, patch

from MEDfl.LearningManager.flpipeline import FLpipeline
from MEDfl.LearningManager.server import FlowerServer


class TestFLpipeline(unittest.TestCase):
    def setUp(self):
        self.name = "test_pipeline"
        self.description = "Testing FLpipeline"
        self.server = Mock(spec=FlowerServer)
        self.pipeline = FLpipeline(self.name, self.description, self.server)

    def test_init(self):
        self.assertEqual(self.pipeline.name, self.name)
        self.assertEqual(self.pipeline.description, self.description)
        self.assertEqual(self.pipeline.server, self.server)

    def test_validate(self):
        # Test valid attributes
        self.pipeline.validate()

        # Test invalid name
        self.pipeline.name = 123
        with self.assertRaises(TypeError):
            self.pipeline.validate()
        self.pipeline.name = self.name

        # Test invalid description
        self.pipeline.description = 123
        with self.assertRaises(TypeError):
            self.pipeline.validate()
        self.pipeline.description = self.description

        # Test invalid server
        self.pipeline.server = "invalid_server"
        with self.assertRaises(TypeError):
            self.pipeline.validate()
        self.pipeline.server = self.server

    def test_create(self):
        result = "Test Result"
        creation_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        with patch("your_package.FLpipeline.my_eng.execute") as mock_execute:
            self.pipeline.create(result)

            mock_execute.assert_called_once_with(
                f"INSERT INTO FLpipeline(name, description, creation_date, results) "
                f"VALUES ('{self.name}', '{self.description}', '{creation_date}', '{result}')"
            )

            # Assuming the following line is executed based on successful execution
            self.server.fed_dataset.update.assert_called_once()

    def test_update(self):
        # Placeholder test for update method
        self.pipeline.update()
        # You can add specific test cases or assertions here if needed

    def test_delete(self):
        # Placeholder test for delete method
        with patch("your_package.FLpipeline.my_eng.execute") as mock_execute:
            self.pipeline.delete()

            mock_execute.assert_called_once_with(
                f"DELETE FROM FLpipeline WHERE name = '{self.name}'"
            )

    def test_test_by_node(self):
        node_name = "test_node"
        test_frac = 0.5
        expected_result = {
            "node_name": node_name,
            "classification_report": "Mock Classification Report",
        }
        self.server.fed_dataset.test_nodes.index.return_value = 0
        self.server.global_model.model = Mock()
        self.server.fed_dataset.testloaders[0] = Mock()
        self.server.fed_dataset.testloaders[0].dataset = Mock()
        self.server.fed_dataset.testloaders[
            0
        ].dataset.__getitem__.side_effect = [
            ("data1", "target1"),
            ("data2", "target2"),
        ]
        self.server.global_model.model.return_value = "Mock Model"

        with patch("your_package.FLpipeline.test") as mock_test:
            mock_test.return_value = "Mock Classification Report"
            result = self.pipeline.test_by_node(node_name, test_frac)

            mock_test.assert_called_once_with(
                model="Mock Model", test_loader=...
            )
            self.assertEqual(result, expected_result)

    def test_auto_test(self):
        node_names = ["test_node1", "test_node2"]
        test_frac = 0.5
        expected_result = [
            {
                "node_name": node_names[0],
                "classification_report": "Mock Classification Report",
            },
            {
                "node_name": node_names[1],
                "classification_report": "Mock Classification Report",
            },
        ]
        self.server.fed_dataset.test_nodes = node_names

        with patch.object(self.pipeline, "test_by_node") as mock_test_by_node:
            mock_test_by_node.side_effect = [
                "Mock Classification Report"
            ] * len(node_names)
            result = self.pipeline.auto_test(test_frac)

            self.assertEqual(mock_test_by_node.call_count, len(node_names))
            self.assertEqual(result, expected_result)


if __name__ == "__main__":
    unittest.main()



from MEDfl.NetManager.net_helper import *
from MEDfl.NetManager.net_manager_queries import *
from MEDfl.NetManager.database_connector import DatabaseManager

class FederatedDataset:
    def __init__(
        self,
        name: str,
        train_nodes: list,
        test_nodes: list,
        trainloaders: list,
        valloaders: list,
        testloaders: list,
    ):
        """
        Represents a Federated Dataset.

        :param name: Name of the Federated Dataset.
        :param train_nodes: List of train nodes.
        :param test_nodes: List of test nodes.
        :param trainloaders: List of train data loaders.
        :param valloaders: List of validation data loaders.
        :param testloaders: List of test data loaders.
        """
        self.name = name
        self.train_nodes = train_nodes
        self.test_nodes = test_nodes
        self.trainloaders = trainloaders
        self.valloaders = valloaders
        self.testloaders = testloaders
        self.size = len(self.trainloaders[0].dataset[0][0])

        db_manager = DatabaseManager()
        db_manager.connect()
        self.eng = db_manager.get_connection()

    def create(self, FLsetupId: int):
        """
        Create a new Federated Dataset in the database.

        :param FLsetupId: The FLsetup ID associated with the Federated Dataset.
        """
        query_params = {"name": self.name, "FLsetupId": FLsetupId}
        fedDataId = get_feddataset_id_from_name(self.name)
        if fedDataId :
            self.id = fedDataId
        else:
            self.eng.execute(text(INSERT_FLDATASET_QUERY), query_params)
            self.id = get_feddataset_id_from_name(self.name)


    def update(self, FLpipeId: int, FedId: int):
        """
        Update the FLpipe ID associated with the Federated Dataset in the database.

        :param FLpipeId: The new FLpipe ID to be updated.
        :param FedId: The Federated Dataset ID.
        """
        query_params = {"FLpipeId": FLpipeId, "FedId": FedId}
        self.eng.execute(text(UPDATE_FLDATASET_QUERY), **query_params)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader
from sklearn.model_selection import GridSearchCV, train_test_split
from sklearn.base import BaseEstimator
from sklearn.metrics import make_scorer, precision_score, recall_score, accuracy_score, f1_score,roc_auc_score, balanced_accuracy_score
import optuna

from MEDfl.LearningManager.model import Model
from MEDfl.LearningManager.strategy import Strategy
from MEDfl.LearningManager.server import FlowerServer
from MEDfl.LearningManager.flpipeline import FLpipeline

class BinaryClassifier(nn.Module):
    def __init__(self, input_size, num_layers, layer_size):
        super(BinaryClassifier, self).__init__()

        # Input layer
        self.layers = [nn.Linear(input_size, layer_size)]
        
        # Hidden layers
        for _ in range(num_layers - 1):
            self.layers.append(nn.Linear(layer_size, layer_size))
        
        # Output layer
        self.layers.append(nn.Linear(layer_size, 1))
        
        # ModuleList to handle dynamic number of layers
        self.layers = nn.ModuleList(self.layers)

    def forward(self, x):
        for layer in self.layers[:-1]:
            x = F.relu(layer(x))
        x = self.layers[-1](x)
        return x
    
class CustomPyTorchClassifier(BaseEstimator):
    def __init__(self, hidden_dim=10, lr=0.001, pos_weight=1, th=0.5, max_epochs=10, batch_size=32):
        self.hidden_dim = hidden_dim
        self.lr = lr
        self.pos_weight = pos_weight
        self.max_epochs = max_epochs
        self.batch_size = batch_size
        self.th = th
        self.model = None

    def fit(self, X, y):
        if isinstance(X, torch.Tensor):
            X = X.numpy()
        if isinstance(y, torch.Tensor):
            y = y.numpy()

        input_dim = X.shape[1]
        self.model = nn.Sequential(
            nn.Linear(input_dim, self.hidden_dim),
            nn.ReLU(),
            nn.Linear(self.hidden_dim, 1),
            nn.Sigmoid()
        )

        criterion = nn.BCEWithLogitsLoss(pos_weight=torch.tensor(self.pos_weight))
        optimizer = optim.Adam(self.model.parameters(), lr=self.lr)

        train_data = TensorDataset(torch.from_numpy(X).float(), torch.from_numpy(y).float())
        train_loader = DataLoader(train_data, batch_size=self.batch_size, shuffle=True)

        self.model.train()
        for epoch in range(self.max_epochs):
            for inputs, labels in train_loader:
                optimizer.zero_grad()
                outputs = self.model(inputs)
                loss = criterion(outputs.squeeze(), labels)
                loss.backward()
                optimizer.step()
        return self

    def predict(self, X):
        if isinstance(X, torch.Tensor):
            X = X.numpy()

        self.model.eval()
        with torch.no_grad():
            outputs = self.model(torch.from_numpy(X).float())
            predictions = (outputs.squeeze() > self.th).float().numpy()
        return predictions

    def score(self, X, y):
        predictions = self.predict(X)
        return accuracy_score(y, predictions)


class ParamsOptimiser:
    def __init__(self, X_train = None, y_train=None, X_test=None, y_test=None):
        if isinstance(X_train, pd.DataFrame):
            X_train = X_train.to_numpy()
        if isinstance(y_train, pd.Series):
            y_train = y_train.to_numpy()
        if isinstance(X_test, pd.DataFrame):
            X_test = X_test.to_numpy()
        if isinstance(y_test, pd.Series):
            y_test = y_test.to_numpy()

        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test

    def perform_grid_search(self, param_grid, scoring_metric='recall', cv=3, verbose=1):
        pytorch_model = CustomPyTorchClassifier()
        scorer = make_scorer(recall_score, greater_is_better=True)

        if scoring_metric == 'precision':
            scorer = make_scorer(precision_score)
        elif scoring_metric == 'accuracy':
            scorer = make_scorer(accuracy_score)
        elif scoring_metric == 'recall':
            scorer = make_scorer(recall_score)
        elif scoring_metric == 'f1':
            scorer = make_scorer(f1_score)

        grid_search = GridSearchCV(pytorch_model, param_grid, scoring=scorer, cv=cv, refit=scoring_metric, verbose=verbose)
        grid_search.fit(self.X_train, self.y_train)

        self.grid_search_results = grid_search  # Save the grid search results

        return grid_search
    
    # Inside the CustomModelTrainer class
    def plot_results(self, params_to_plot=None):
        results = pd.DataFrame(self.grid_search_results.cv_results_)

        if params_to_plot is None:
            # Create a column for configuration details
            results['config'] = results['params'].apply(lambda x: str(x))

            # Visualize mean test scores along with configurations
            plt.figure(figsize=(15, 8))
            bar_plot = plt.bar(results.index, results['mean_test_score'], color='blue', alpha=0.7)
            plt.xticks(results.index, results['config'], rotation='vertical', fontsize=8)
            plt.ylabel('Mean Test Score')
            plt.title('Mean Test Scores for Each Configuration')
            plt.tight_layout()

            # Add values on top of bars
            for bar, score in zip(bar_plot, results['mean_test_score']):
                plt.text(bar.get_x() + bar.get_width() / 2 - 0.15, bar.get_height() + 0.01, f'{score:.3f}', fontsize=8)

            plt.show()
            return 

        try:
            # Dynamically get the column names for the specified scoring metric
            mean_test_col = f'mean_test_{params_to_plot[0]}'
            param_cols = [f'param_{param}' for param in params_to_plot]

            if len(params_to_plot) == 1:
                # Plotting the heatmap for a single parameter
                plt.figure(figsize=(8, 6))
                sns.heatmap(results.pivot_table(index=param_cols[0]),
                            annot=True, cmap='YlGnBu', fmt=".3f", cbar_kws={'label': mean_test_col})
                plt.title(mean_test_col.capitalize())
                plt.show()
            elif len(params_to_plot) == 2:
                # Create a pair plot for two parameters
                plt.figure(figsize=(8, 6))
                scores = results.pivot_table(index=param_cols[0], columns=param_cols[1], values=f'mean_test_score', aggfunc="mean")
                sns.heatmap(scores, annot=True, cmap='YlGnBu', fmt=".3f", cbar_kws={'label': mean_test_col})
                plt.title(mean_test_col.capitalize())
                plt.show()
            else:
                print("Invalid number of parameters to plot. You can provide either one or two parameters.")
        except KeyError as e:
            print(f"Error: {e}. Make sure the specified scoring metric exists in the results DataFrame.")

    

    def optuna_optimisation(self, direction, params):
        # Create the data loaders here
        train_data = TensorDataset(torch.from_numpy(self.X_train).float(), torch.from_numpy(self.y_train).float())
        test_data = TensorDataset(torch.from_numpy(self.X_test).float(), torch.from_numpy(self.y_test).float())

       

        def objective(trial):

            batch_size=trial.suggest_int('batch_size', **params['batch_size'])

            train_loader = DataLoader(train_data, batch_size=batch_size, shuffle=True)
            test_loader = DataLoader(test_data, batch_size=batch_size, shuffle=False)

            # Create the model with the suggested hyperparameters
            model = BinaryClassifier(input_size=self.X_train.shape[1],
                                     num_layers=trial.suggest_int('num_layers', **params['num_layers']) ,
                                     layer_size=trial.suggest_int('hidden_size', **params['hidden_size']))

            # Define the loss function and optimizer
            criterion = nn.BCEWithLogitsLoss()
            optimizer_name = trial.suggest_categorical('optimizer', params['optimizer'])
            learning_rate = trial.suggest_float('learning_rate', **params['learning_rate'])
            

            if optimizer_name == 'Adam':
                optimizer = optim.Adam(model.parameters(), lr=learning_rate)
            elif optimizer_name == 'SGD':
                optimizer = optim.SGD(model.parameters(), lr=learning_rate)
            elif optimizer_name == 'RMSprop':
                optimizer = optim.RMSprop(model.parameters(), lr=learning_rate)

            # Training loop
            num_epochs = trial.suggest_int('num_epochs', **params['num_epochs'])
            for epoch in range(num_epochs):
                model.train()
                for batch_X, batch_y in train_loader:
                    optimizer.zero_grad()
                    outputs = model(batch_X)
                    loss = criterion(outputs.squeeze(), batch_y)
                    loss.backward()
                    optimizer.step()

                # Evaluation
                model.eval()
                predictions = []
                true_labels = []
                with torch.no_grad():
                    for batch_X, batch_y in test_loader:
                        outputs = model(batch_X)
                        predictions.extend(torch.sigmoid(outputs).numpy())
                        true_labels.extend(batch_y.numpy())

                # Calculate F1 score
                # f1 = f1_score(true_labels, (np.array(predictions) > 0.5).astype(int))
                auc = roc_auc_score(true_labels, predictions)

                trial.report(auc, epoch)

                # Handle pruning based on the intermediate value
                if trial.should_prune():
                    raise optuna.TrialPruned()

            return auc

        # Create an Optuna study
        study = optuna.create_study(direction=direction)
        study.optimize(objective, n_trials=params['n_trials'])
        
        self.study = study

        # Get the best hyperparameters
        best_params = study.best_params
        print(f"Best Hyperparameters: {best_params}")

        return study
    
    def train_optimized_model(self ,trial ,th_min , th_max):
        
        best_params = self.study.best_params

        threshold = trial.suggest_float('threashhold', th_min, th_max, log=True)

        train_data = TensorDataset(torch.from_numpy(self.X_train).float(), torch.from_numpy(self.y_train).float())
        test_data = TensorDataset(torch.from_numpy(self.X_test).float(), torch.from_numpy(self.y_test).float())

        train_loader = DataLoader(train_data, batch_size=best_params['batch_size'], shuffle=True)
        test_loader = DataLoader(test_data, batch_size=best_params['batch_size'], shuffle=False)

        
        # Use the best hyperparameters to train the final model
        final_model = BinaryClassifier(input_size=self.X_train.shape[1], layer_size=best_params['hidden_size'] , num_layers=best_params['num_layers'])
        final_optimizer = self.get_optimizer(best_params['optimizer'], final_model.parameters(), best_params['learning_rate'])
        final_criterion = nn.BCEWithLogitsLoss()

        num_epochs = best_params['num_epochs']
        for epoch in range(num_epochs):
            final_model.train()
            for batch_X, batch_y in train_loader:
                final_optimizer.zero_grad()
                outputs = final_model(batch_X)
                loss = final_criterion(outputs.squeeze(), batch_y)
                loss.backward()
                final_optimizer.step()

        # Evaluate the final model on the test set
        final_model.eval()
        with torch.no_grad():
            predictions = []
            true_labels = []
            for batch_X, batch_y in test_loader:
                outputs = final_model(batch_X)
                predictions.extend(torch.sigmoid(outputs).numpy())
                true_labels.extend(batch_y.numpy())

        final_balanced_acc = balanced_accuracy_score(true_labels, (np.array(predictions) > threshold).astype(int))
        print(f"Model balanced accuracy: {final_balanced_acc}")

        return final_balanced_acc

    def get_optimizer(self, optimizer_name, parameters, learning_rate):
        if optimizer_name == 'Adam':
            return optim.Adam(parameters, lr=learning_rate)
        elif optimizer_name == 'SGD':
            return optim.SGD(parameters, lr=learning_rate)
        elif optimizer_name == 'RMSprop':
            return optim.RMSprop(parameters, lr=learning_rate)
        else:
            raise ValueError(f"Unknown optimizer: {optimizer_name}")

    def perform_grid_search(self, param_grid, scoring_metric='recall', cv=3, verbose=1):
        pytorch_model = CustomPyTorchClassifier()
        scorer = make_scorer(recall_score, greater_is_better=True)

        if scoring_metric == 'precision':
            scorer = make_scorer(precision_score)
        elif scoring_metric == 'accuracy':
            scorer = make_scorer(accuracy_score)
        elif scoring_metric == 'recall':
            scorer = make_scorer(recall_score)
        elif scoring_metric == 'f1':
            scorer = make_scorer(f1_score)

        grid_search = GridSearchCV(pytorch_model, param_grid, scoring=scorer, cv=cv, refit=scoring_metric, verbose=verbose)
        grid_search.fit(self.X_train, self.y_train)

        self.grid_search_results = grid_search  # Save the grid search results

        return grid_search
    
    
    def plot_param_importances(self):
        return optuna.visualization.plot_param_importances(self.study)
    
    def plot_slice(self , params):
        return optuna.visualization.plot_slice(self.study , params=params)
    
    def plot_parallel_coordinate(self):
        return optuna.visualization.plot_parallel_coordinate(self.study)
    
    def plot_rank(self , params=None):
        return optuna.visualization.plot_rank(self.study , params=params)
    
    def plot_optimization_history(self):
        return optuna.visualization.plot_optimization_history(self.study)
    
    def optimize_model_threashhold(self , n_trials , th_min , th_max):
        additional_params = {'th_min': th_min, 'th_max': th_max}

        th_study = optuna.create_study(direction='maximize')
        th_study.optimize(lambda trial: self.train_optimized_model(trial , **additional_params) , n_trials)

        # Get the best hyperparameters
        best_params = th_study.best_params
        print(f"Best Hyperparameters: {best_params}")

        return optuna.visualization.plot_rank(th_study , params=['threashhold'])
    
    def federated_params_iptim(self , params , direction,  model, fl_dataset):

        def objective(trial):
          
            criterion = nn.BCEWithLogitsLoss()

            optimizer_name = trial.suggest_categorical('optimizer', params['optimizer'])
            learning_rate = trial.suggest_float('learning_rate', **params['learning_rate'])
            num_rounds = trial.suggest_int('num_rounds', **params['num_rounds'])
            diff_privacy =  trial.suggest_int('diff_privacy', **params['diff_privacy'])
            diff_privacy = True if diff_privacy == 1 else False

            if optimizer_name == 'Adam':
                optimizer = optim.Adam(model.parameters(), lr=learning_rate)
            elif optimizer_name == 'SGD':
                optimizer = optim.SGD(model.parameters(), lr=learning_rate)
            elif optimizer_name == 'RMSprop':
                optimizer = optim.RMSprop(model.parameters(), lr=learning_rate)

            # Creating a new Model instance using the specific model created by DynamicModel
            global_model = Model(model, optimizer, criterion)

            # Get the initial params of the model 
            init_params = global_model.get_parameters() 

            fl_strategy = trial.suggest_categorical('fl_strategy', params['fl_strategy'])

            learning_strategy = Strategy(fl_strategy, 
                   fraction_fit = 1.0 ,
                   fraction_evaluate = 1.0,
                   min_fit_clients = 2,
                   min_evaluate_clients = 2,
                   min_available_clients = 2 , 
                   initial_parameters=init_params)
            
            learning_strategy.create_strategy()

            # Create The server 
            server = FlowerServer(global_model, strategy = learning_strategy, num_rounds = num_rounds,
                       num_clients  = len(fl_dataset.trainloaders), 
                       fed_dataset = fl_dataset,diff_privacy = diff_privacy,
                       # You can change the resources alocated for each client based on your machine 
                       client_resources={'num_cpus': 1.0, 'num_gpus': 0.0}
                       )
            
            ppl_1 = FLpipeline( name ="the first fl_pipeline",description = "this is our first FL pipeline",
                   server = server)
            
            # Run the Traning of the model
            history = ppl_1.server.run()

            return server.auc[len(server.auc)-1]
        

       
        study = optuna.create_study(direction=direction)
        study.optimize(objective, n_trials=params['n_trials'])
        
        self.study = study

        # Get the best hyperparameters
        best_params = study.best_params
        print(f"Best Hyperparameters: {best_params}")

        return study


        








        
    







#!/usr/bin/env python3
import flwr as fl
from opacus import PrivacyEngine
from torch.utils.data import DataLoader

from .model import Model
from .utils import params
import torch

class FlowerClient(fl.client.NumPyClient):
    """
    FlowerClient class for creating MEDfl clients.

    Attributes:
        cid (str): Client ID.
        local_model (Model): Local model of the federated learning network.
        trainloader (DataLoader): DataLoader for training data.
        valloader (DataLoader): DataLoader for validation data.
        diff_priv (bool): Flag indicating whether to use differential privacy.
    """
    def __init__(self, cid: str, local_model: Model, trainloader: DataLoader, valloader: DataLoader, diff_priv: bool = params["diff_privacy"]):
        """
        Initializes the FlowerClient instance.

        Args:
            cid (str): Client ID.
            local_model (Model): Local model of the federated learning network.
            trainloader (DataLoader): DataLoader for training data.
            valloader (DataLoader): DataLoader for validation data.
            diff_priv (bool): Flag indicating whether to use differential privacy.
        """
        self.cid = cid
        self.local_model = local_model
        self.trainloader = trainloader
        self.valloader = valloader
        if torch.cuda.is_available():
            num_cuda_devices = torch.cuda.device_count()
            if num_cuda_devices > 0:
                device_idx = int(self.cid) % num_cuda_devices
                self.device = torch.device(f"cuda:{device_idx}")
                self.local_model.model.to(self.device)
            else:
                # Handle case where CUDA is available but no CUDA devices are found
                raise RuntimeError("CUDA is available, but no CUDA devices are found.")
        else:
            # Handle case where CUDA is not available
            self.device = torch.device("cpu")
            self.local_model.model.to(self.device)

        self.privacy_engine = PrivacyEngine(secure_mode=False)
        self.diff_priv = diff_priv
        self.epsilons = []
        self.accuracies = []
        self.losses = []
        if self.diff_priv:
            model, optimizer, self.trainloader = self.privacy_engine.make_private_with_epsilon(
                module=self.local_model.model.train(),
                optimizer=self.local_model.optimizer,
                data_loader=self.trainloader,
                epochs=params["train_epochs"],
                target_epsilon=float(params["EPSILON"]),
                target_delta= float(params["DELTA"]),
                max_grad_norm=params["MAX_GRAD_NORM"],
            )
            setattr(self.local_model, "model", model)
            setattr(self.local_model, "optimizer", optimizer)
        self.validate()

    def validate(self):
        """Validates cid, local_model, trainloader, valloader."""
        if not isinstance(self.cid, str):
            raise TypeError("cid argument must be a string")

        if not isinstance(self.local_model, Model):
            raise TypeError("local_model argument must be a MEDfl.LearningManager.model.Model")

        if not isinstance(self.trainloader, DataLoader):
            raise TypeError("trainloader argument must be a torch.utils.data.dataloader")

        if not isinstance(self.valloader, DataLoader):
            raise TypeError("valloader argument must be a torch.utils.data.dataloader")

        if not isinstance(self.diff_priv, bool):
            raise TypeError("diff_priv argument must be a bool")

    def get_parameters(self, config):
        """
        Returns the current parameters of the local model.

        Args:
            config: Configuration information.

        Returns:
            Numpy array: Parameters of the local model.
        """
        print(f"[Client {self.cid}] get_parameters")
        return self.local_model.get_parameters()

    def fit(self, parameters, config):
        """
        Fits the local model to the received parameters using federated learning.

        Args:
            parameters: Parameters received from the server.
            config: Configuration information.

        Returns:
            Tuple: Parameters of the local model, number of training examples, and privacy information.
        """
        print('\n -------------------------------- \n  this is the config of the client')
        print(f"[Client {self.cid}] fit, config: {config}")
        # print(config['epochs'])
        print('\n -------------------------------- \n  ')
        self.local_model.set_parameters(parameters)
        for _ in range(params["train_epochs"]):
            epsilon = self.local_model.train(
                self.trainloader,
                epoch=_,
                device=self.device,
                privacy_engine=self.privacy_engine,
                diff_priv=self.diff_priv,
            )
            self.epsilons.append(epsilon)
        print(f"epsilon of client {self.cid} : eps = {epsilon}")
        return (
            self.local_model.get_parameters(),
            len(self.trainloader),
            {"epsilon": epsilon},
        )

    def evaluate(self, parameters, config):
        """
        Evaluates the local model on the validation data and returns the loss and accuracy.

        Args:
            parameters: Parameters received from the server.
            config: Configuration information.

        Returns:
            Tuple: Loss, number of validation examples, and accuracy information.
        """
        print(f"[Client {self.cid}] evaluate, config: {config}")
        self.local_model.set_parameters(parameters)
        loss, accuracy , auc = self.local_model.evaluate(
            self.valloader, device=self.device
        )
        self.losses.append(loss)
        self.accuracies.append(accuracy)
        
        return float(loss), len(self.valloader), {"accuracy": float(accuracy)}



from datetime import datetime


from torch.utils.data import random_split, DataLoader, Dataset

from MEDfl.LearningManager.federated_dataset import FederatedDataset
from .net_helper import *
from .net_manager_queries import *  # Import the sql_queries module
from .network import Network

from .node import Node

from MEDfl.NetManager.database_connector import DatabaseManager


class FLsetup:
    def __init__(self, name: str, description: str, network: Network):
        """Initialize a Federated Learning (FL) setup.

        Args:
            name (str): The name of the FL setup.
            description (str): A description of the FL setup.
            network (Network): An instance of the Network class representing the network architecture.
        """
        self.name = name
        self.description = description
        self.network = network
        self.column_name = None
        self.auto = 1 if self.column_name is not None else 0
        self.validate()
        self.fed_dataset = None

        db_manager = DatabaseManager()
        db_manager.connect()
        self.eng = db_manager.get_connection()

        

    def validate(self):
        """Validate name, description, and network."""
        if not isinstance(self.name, str):
            raise TypeError("name argument must be a string")

        if not isinstance(self.description, str):
            raise TypeError("description argument must be a string")

        if not isinstance(self.network, Network):
            raise TypeError(
                "network argument must be a MEDfl.NetManager.Network "
            )

    def create(self):
        """Create an FL setup."""
        creation_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        netid = get_netid_from_name(self.network.name)
        self.eng.execute(
            text(CREATE_FLSETUP_QUERY),
            {
                "name": self.name,
                "description": self.description,
                "creation_date": creation_date,
                "net_id": netid,
                "column_name": self.column_name,
            },
        )
        self.id = get_flsetupid_from_name(self.name)

    def delete(self):
        """Delete the FL setup."""
        if self.fed_dataset is not None:
            self.fed_dataset.delete_Flsetup(FLsetupId=self.id)
        self.eng.execute(text(DELETE_FLSETUP_QUERY), {"name": self.name})

    @classmethod
    def read_setup(cls, FLsetupId: int):
        """Read the FL setup by FLsetupId.

        Args:
            FLsetupId (int): The id of the FL setup to read.

        Returns:
            FLsetup: An instance of the FLsetup class with the specified FLsetupId.
        """
        db_manager = DatabaseManager()
        db_manager.connect()
        my_eng = db_manager.get_connection()

        res = pd.read_sql(
            text(READ_SETUP_QUERY), my_eng, params={"flsetup_id": FLsetupId}
        ).iloc[0]
        
        network_res = pd.read_sql(
            text(READ_NETWORK_BY_ID_QUERY),
            my_eng,
            params={"net_id": int(res["NetId"])},
        ).iloc[0]
        network = Network(network_res["NetName"])
        setattr(network, "id", res["NetId"])
        fl_setup = cls(res["name"], res["description"], network)
        if res["column_name"] == str(None):
            res["column_name"] = None
        setattr(fl_setup, "column_name", res["column_name"])
        setattr(fl_setup, "id", res["FLsetupId"])

        return fl_setup

    @staticmethod
    def list_allsetups():
        """List all the FL setups.

        Returns:
            DataFrame: A DataFrame containing information about all the FL setups.
        """
        db_manager = DatabaseManager()
        db_manager.connect()
        my_eng = db_manager.get_connection()

        Flsetups = pd.read_sql(text(READ_ALL_SETUPS_QUERY), my_eng)
        return Flsetups

    def create_nodes_from_master_dataset(self, params_dict: dict):
        """Create nodes from the master dataset.

        Args:
            params_dict (dict): A dictionary containing parameters for node creation.
                - column_name (str): The name of the column in the MasterDataset used to create nodes.
                - train_nodes (list): A list of node names that will be used for training.
                - test_nodes (list): A list of node names that will be used for testing.

        Returns:
            list: A list of Node instances created from the master dataset.
        """
        assert "column_name" in params_dict.keys()
        column_name, train_nodes, test_nodes = (
            params_dict["column_name"],
            params_dict["train_nodes"],
            params_dict["test_nodes"],
        )
        self.column_name = column_name
        self.auto = 1

        # Update the Column name of the auto flSetup
        query = f"UPDATE FLsetup SET column_name = '{column_name}' WHERE name = '{self.name}'"
        self.eng.execute(text(query))
        

        # Add Network to DB
        # self.network.create_network()

        netid = get_netid_from_name(self.network.name)

        assert self.network.mtable_exists == 1
        node_names = pd.read_sql(
            text(READ_DISTINCT_NODES_QUERY.format(column_name)), self.eng
        )

        nodes = [Node(val[0], 1) for val in node_names.values.tolist()]

        used_nodes = []

        for node in nodes:
            if node.name in train_nodes:
                node.train = 1
                node.create_node(netid)
                used_nodes.append(node)
            if node.name in test_nodes:
                node.train =0
                node.create_node(netid) 
                used_nodes.append(node)
        return used_nodes

    def create_dataloader_from_node(
        self,
        node: Node,
        output,
        fill_strategy="mean",  fit_encode=[], to_drop=[],
        train_batch_size: int = 32,
        test_batch_size: int = 1,
        split_frac: float = 0.2,
        dataset: Dataset = None,

    ):
        """Create DataLoader from a Node.

        Args:
            node (Node): The node from which to create DataLoader.
            train_batch_size (int): The batch size for training data.
            test_batch_size (int): The batch size for test data.
            split_frac (float): The fraction of data to be used for training.
            dataset (Dataset): The dataset to use. If None, the method will read the dataset from the node.

        Returns:
            DataLoader: The DataLoader instances for training and testing.
        """
        if dataset is None:
            if self.column_name is not None:
                dataset = process_data_after_reading(
                    node.get_dataset(self.column_name), output, fill_strategy=fill_strategy, fit_encode=fit_encode, to_drop=to_drop
                )
            else:
                dataset = process_data_after_reading(
                    node.get_dataset(), output, fill_strategy=fill_strategy, fit_encode=fit_encode, to_drop=to_drop)

        dataset_size = len(dataset)
        traindata_size = int(dataset_size * (1 - split_frac))
        traindata, testdata = random_split(
            dataset, [traindata_size, dataset_size - traindata_size]
        )
        trainloader, testloader = DataLoader(
            traindata, batch_size=train_batch_size
        ), DataLoader(testdata, batch_size=test_batch_size)
        return trainloader, testloader

    def create_federated_dataset(
        self, output, fill_strategy="mean",  fit_encode=[], to_drop=[], val_frac=0.1, test_frac=0.2
    ) -> FederatedDataset:
        """Create a federated dataset.

        Args:
            output (string): the output feature of the dataset
            val_frac (float): The fraction of data to be used for validation.
            test_frac (float): The fraction of data to be used for testing.

        Returns:
            FederatedDataset: The FederatedDataset instance containing train, validation, and test data.
        """
        
        if not self.column_name:
            to_drop.extend(["DataSetName" , "NodeId" , "DataSetId"])
        else :
            to_drop.extend(["PatientId"]) 
            
        netid = self.network.id
        train_nodes = pd.read_sql(
            text(
                f"SELECT Nodes.NodeName  FROM Nodes WHERE Nodes.NetId = {netid} AND Nodes.train = 1 "
            ),
            self.eng,
        )
        test_nodes = pd.read_sql(
            text(
                f"SELECT Nodes.NodeName  FROM Nodes WHERE Nodes.NetId = {netid} AND Nodes.train = 0 "
            ),
            self.eng,
        )

        train_nodes = [
            Node(val[0], 1, test_frac) for val in train_nodes.values.tolist()
        ]
        test_nodes = [Node(val[0], 0) for val in test_nodes.values.tolist()]

        trainloaders, valloaders, testloaders = [], [], []
        # if len(test_nodes) == 0:
        #     raise "test node empty"
        if test_nodes is None:
            _, testloader = self.create_dataloader_from_node(
                train_nodes[0], output, fill_strategy=fill_strategy, fit_encode=fit_encode, to_drop=to_drop)
            testloaders.append(testloader)
        else:
            for train_node in train_nodes:
                train_valloader, testloader = self.create_dataloader_from_node(
                    train_node, output, fill_strategy=fill_strategy,
                    fit_encode=fit_encode, to_drop=to_drop,)
                trainloader, valloader = self.create_dataloader_from_node(
                    train_node,
                    output, fill_strategy=fill_strategy, fit_encode=fit_encode, to_drop=to_drop,
                    test_batch_size=32,
                    split_frac=val_frac,
                    dataset=train_valloader.dataset,
                )
                trainloaders.append(trainloader)
                valloaders.append(valloader)
                testloaders.append(testloader)

            for test_node in test_nodes:
                _, testloader = self.create_dataloader_from_node(
                    test_node, output, fill_strategy=fill_strategy, fit_encode=fit_encode, to_drop=to_drop, split_frac=1.0
                )
                testloaders.append(testloader)
        train_nodes_names = [node.name for node in train_nodes]
        test_nodes_names = train_nodes_names + [
            node.name for node in test_nodes
        ]
        
        # test_nodes_names = [
        #     node.name for node in test_nodes
        # ]

        # Add FlSetup on to the DataBase
        # self.create()

        # self.network.update_network(FLsetupId=self.id)
        fed_dataset = FederatedDataset(
            self.name + "_Feddataset",
            train_nodes_names,
            test_nodes_names,
            trainloaders,
            valloaders,
            testloaders,
        )
        self.fed_dataset = fed_dataset
        self.fed_dataset.create(self.id)
        return self.fed_dataset
    
    


    def get_flDataSet(self):
        """
        Retrieve the federated dataset associated with the FL setup using the FL setup's name.
 
        Returns:
            pandas.DataFrame: DataFrame containing the federated dataset information.
        """
        return pd.read_sql(
            text(
                f"SELECT * FROM FedDatasets WHERE FLsetupId = {get_flsetupid_from_name(self.name)}"
            ),
            self.eng,
        )




from collections import OrderedDict
from typing import Dict, List, Optional, Tuple

import flwr as fl
import numpy as np

import optuna




class  Strategy:
    """
    A class representing a strategy for Federated Learning.

    Attributes:
        name (str): The name of the strategy. Default is "FedAvg".
        fraction_fit (float): Fraction of clients to use for training during each round. Default is 1.0.
        fraction_evaluate (float): Fraction of clients to use for evaluation during each round. Default is 1.0.
        min_fit_clients (int): Minimum number of clients to use for training during each round. Default is 2.
        min_evaluate_clients (int): Minimum number of clients to use for evaluation during each round. Default is 2.
        min_available_clients (int): Minimum number of available clients required to start a round. Default is 2.
        initial_parameters (Optional[]): The initial parameters of the server model 
    Methods:
     
    """

    def __init__(
        self,
        name: str = "FedAvg",
        fraction_fit: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_fit_clients: int = 2,
        min_evaluate_clients: int = 2,
        min_available_clients: int = 2,
        initial_parameters = [],
        evaluation_methode = "centralized"
    ) -> None:
        """
        Initialize a Strategy object with the specified parameters.

        Args:
            name (str): The name of the strategy. Default is "FedAvg".
            fraction_fit (float): Fraction of clients to use for training during each round. Default is 1.0.
            fraction_evaluate (float): Fraction of clients to use for evaluation during each round. Default is 1.0.
            min_fit_clients (int): Minimum number of clients to use for training during each round. Default is 2.
            min_evaluate_clients (int): Minimum number of clients to use for evaluation during each round. Default is 2.
            min_available_clients (int): Minimum number of available clients required to start a round. Default is 2.
            initial_parameters (Optional[]): The initial parametres of the server model 
            evaluation_methode ( "centralized" | "distributed")
        """
        self.fraction_fit = fraction_fit
        self.fraction_evaluate = fraction_evaluate
        self.min_fit_clients = min_fit_clients
        self.min_evaluate_clients = min_evaluate_clients
        self.min_available_clients = min_available_clients
        self.initial_parameters = initial_parameters
        self.evaluate_fn = None
        self.name = name
    
    def optuna_fed_optimization(self, direction:str , hpo_rate:int , params_config):
        self.study = optuna.create_study(direction=direction)
        self.hpo_rate = hpo_rate
        self.params_config = params_config
    
 
    def create_strategy(self):
        self.strategy_object = self.get_strategy_by_name()(
            fraction_fit=self.fraction_fit,
            fraction_evaluate=self.fraction_evaluate,
            min_fit_clients=self.min_fit_clients,
            min_evaluate_clients=self.min_evaluate_clients,
            min_available_clients=self.min_available_clients,
            initial_parameters=fl.common.ndarrays_to_parameters(self.initial_parameters),
            evaluate_fn=self.evaluate_fn
        )
    def get_strategy_by_name(self):
        return eval(f"fl.server.strategy.{self.name}")
    
    




# MEDfl/LearningManager/__init__.py

# Import modules from this package
from .client import *
from .dynamicModal import *
from .flpipeline import *
from .federated_dataset import *
from .model import *
from .params_optimiser import *
from .plot import *
from .server import *
from .strategy import *
from .utils import *



# # MEDfl/NetworkManager/__init__.py

# # Import modules from this package
from .dataset import *
from .flsetup import *
from .net_helper import *
from .net_manager_queries import *
from .network import *
from .node import *
from .database_connector import *



import unittest

from scripts.base import *
from MEDfl.NetManager.flsetup import FLsetup
from MEDfl.NetManager.network import Network


class TestFLsetup(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Create an in-memory database for testing
        pass

    def test_create_flsetup(self):
        network = Network(name="TestNetwork")
        network.create_network()
        fl_setup = FLsetup(
            name="TestFLsetup", description="Test description", network=network
        )
        fl_setup.create()

        # Verify if the FLsetup is created in the database
        query = text("SELECT * FROM FLsetup WHERE name = 'TestFLsetup'")
        result = my_eng.execute(query).fetchone()
        self.assertIsNotNone(result)
        self.assertEqual(result["name"], "TestFLsetup")
        self.assertEqual(result["description"], "Test description")
        self.assertEqual(result["NetId"], network.id)
        self.assertEqual(result["column_name"], None)

    def test_delete_flsetup(self):
        # Create a test FLsetup in the database
        network = Network(name="TestNetwork")
        network.create_network()
        fl_setup = FLsetup(
            name="TestFLsetup", description="Test description", network=network
        )
        fl_setup.create()

        # Delete the test FLsetup
        fl_setup.delete()

        # Verify if the FLsetup is deleted from the database
        query = text("SELECT * FROM FLsetup WHERE name = 'TestFLsetup'")
        result = my_eng.execute(query).fetchone()
        self.assertIsNone(result)

    def test_read_setup(self):
        # Create a test FLsetup in the database
        network = Network(name="TestNetwork")
        network.create_network()
        fl_setup = FLsetup(
            name="TestFLsetup", description="Test description", network=network
        )
        fl_setup.create()

        # Read the FLsetup from the database using its ID
        fl_setup_id = fl_setup.id
        retrieved_setup = FLsetup.read_setup(fl_setup_id)

        # Verify if the retrieved FLsetup matches the original one
        self.assertEqual(retrieved_setup.name, fl_setup.name)
        self.assertEqual(retrieved_setup.description, fl_setup.description)
        self.assertEqual(retrieved_setup.network.name, fl_setup.network.name)
        self.assertEqual(retrieved_setup.column_name, fl_setup.column_name)

    # Add more test cases as needed


if __name__ == "__main__":
    unittest.main()



import datetime
from typing import List
import json
import pandas as pd


# File: create_query.py
from sqlalchemy import text
from torch.utils.data import DataLoader, TensorDataset
import torch 

from MEDfl.LearningManager.server import FlowerServer
from MEDfl.LearningManager.utils import params, test
from MEDfl.NetManager.net_helper import get_flpipeline_from_name
from MEDfl.NetManager.net_manager_queries import (CREATE_FLPIPELINE_QUERY,
                                                  DELETE_FLPIPELINE_QUERY , CREATE_TEST_RESULTS_QUERY)
from MEDfl.NetManager.database_connector import DatabaseManager

def create_query(name, description, creation_date, result):
    query = text(
        f"INSERT INTO FLpipeline(name, description, creation_date, results) "
        f"VALUES ('{name}', '{description}', '{creation_date}', '{result}')"
    )
    return query



class FLpipeline:
    """
    FLpipeline class for managing Federated Learning pipelines.

    Attributes:
        name (str): The name of the FLpipeline.
        description (str): A description of the FLpipeline.
        server (FlowerServer): The FlowerServer object associated with the FLpipeline.

    Methods:
        __init__(self, name: str, description: str, server: FlowerServer) -> None:
            Initialize FLpipeline with the specified name, description, and server.
  

    """

    def __init__(
        self, name: str, description: str, server: FlowerServer
    ) -> None:
        self.name = name
        self.description = description
        self.server = server
        self.validate()

        db_manager = DatabaseManager()
        db_manager.connect()
        self.eng = db_manager.get_connection()

    def validate(self) -> None:
        """
        Validate the name, description, and server attributes.
        Raises:
            TypeError: If the name is not a string, the description is not a string,
                      or the server is not a FlowerServer object.
        """
        if not isinstance(self.name, str):
            raise TypeError("name argument must be a string")

        if not isinstance(self.description, str):
            raise TypeError("description argument must be a string")

        # if not isinstance(self.server, FlowerServer):
        #     raise TypeError("server argument must be a FlowerServer")

    def create(self, result: str) -> None:
        """
        Create a new FLpipeline entry in the database with the given result.

        Args:
            result (str): The result string to store in the database.

        """
        creation_date = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        query = CREATE_FLPIPELINE_QUERY.format(
            name=self.name,
            description=self.description,
            creation_date=creation_date,
            result=result,
        )
        self.eng.execute(text(query))
        self.id = get_flpipeline_from_name(self.name)
        try:
            self.server.fed_dataset.update(
                FLpipeId=self.id, FedId=self.server.fed_dataset.id
            )
        except:
            pass

    def delete(self) -> None:
        """
        Delete the FLpipeline entry from the database based on its name.

        Note: This is a placeholder method and needs to be implemented based on your specific database setup.

        """
        # Placeholder code for deleting the FLpipeline entry from the database based on the name.
        # You need to implement the actual deletion based on your database setup.
        self.eng.execute(DELETE_FLPIPELINE_QUERY.format(self.name))


    def test_by_node(self, node_name: str, test_frac=1) -> dict:
        """
        Test the FLpipeline by node with the specified test_frac.

        Args:
            node_name (str): The name of the node to test.
            test_frac (float, optional): The fraction of the test data to use. Default is 1.

        Returns:
            dict: A dictionary containing the node name and the classification report.

        """
        idx = self.server.fed_dataset.test_nodes.index(node_name)
        global_model, test_loader = (
            self.server.global_model,
            self.server.fed_dataset.testloaders[idx],
        )
        
        # Move model to GPU if available
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        global_model.model.to(device)
        
        # Prepare test data
        test_data = test_loader.dataset
        num_samples = int(test_frac * len(test_data))
        test_data = TensorDataset(test_data[:num_samples][0].to(device), test_data[:num_samples][1].to(device))
        
        # Create DataLoader for test data
        test_loader = DataLoader(test_data, batch_size=params["test_batch_size"])

        # Perform testing
        classification_report = test(model=global_model.model, test_loader=test_loader, device=device)

        return {
            "node_name": node_name,
            "classification_report": str(classification_report),
        }


    def auto_test(self, test_frac=1) -> List[dict]:
        """
        Automatically test the FLpipeline on all nodes with the specified test_frac.

        Args:
            test_frac (float, optional): The fraction of the test data to use. Default is 1.

        Returns:
            List[dict]: A list of dictionaries containing the node names and the classification reports.

        """
        result = [
            self.test_by_node(node, test_frac)
            for node in self.server.fed_dataset.test_nodes
        ]
        self.create("\n".join(str(res).replace("'", '"') for res in result))

        # stockage des resultats des tests
        for entry in result:
           node_name = entry['node_name']
           classification_report_str = entry['classification_report']

           # Convert the 'classification_report' string to a dictionary
           classification_report_dict = json.loads(classification_report_str.replace("'", "\""))
           try:
               # Insert record into the 'testResults' table
               query = CREATE_TEST_RESULTS_QUERY.format(
                      pipelineId = self.id, 
                      nodeName = node_name , 
                      confusion_matrix = json.dumps(classification_report_dict['confusion matrix']), 
                      accuracy =classification_report_dict['Accuracy'] , 
                      sensivity = classification_report_dict['Sensitivity/Recall'] ,
                      ppv = classification_report_dict['PPV/Precision'] ,
                      npv= classification_report_dict['NPV'] ,
                      f1score= classification_report_dict['F1-score'] ,
                      fpr= classification_report_dict['False positive rate'] ,
                      tpr= classification_report_dict['True positive rate'] 
                )
               self.eng.execute(text(query))
           except Exception as e:
                # This block will catch any other exceptions
                print(f"An unexpected error occurred: {e}")
               
           

        return result



#!/usr/bin/env python3

import copy
from typing import Dict, Optional, Tuple

import flwr as fl
import torch

from .client import FlowerClient
from .federated_dataset import FederatedDataset
from .model import Model
from .strategy import Strategy


class FlowerServer:
    """
    A class representing the central server for Federated Learning using Flower.

    Attributes:
        global_model (Model): The global model that will be federated among clients.
        strategy (Strategy): The strategy used for federated learning, specifying communication and aggregation methods.
        num_rounds (int): The number of federated learning rounds to perform.
        num_clients (int): The number of clients participating in the federated learning process.
        fed_dataset (FederatedDataset): The federated dataset used for training and evaluation.
        diff_priv (bool): Whether differential privacy is used during the federated learning process.
        accuracies (List[float]): A list to store the accuracy of the global model during each round.
        losses (List[float]): A list to store the loss of the global model during each round.
        flower_clients (List[FlowerClient]): A list to store the FlowerClient objects representing individual clients.
     
    """

    def __init__(
        self,
        global_model: Model,
        strategy: Strategy,
        num_rounds: int,
        num_clients: int,
        fed_dataset: FederatedDataset,
        diff_privacy: bool = False,
        client_resources: Optional[Dict[str, float]] = {'num_cpus': 1, 'num_gpus': 0.0}
    ) -> None:
        """
        Initialize a FlowerServer object with the specified parameters.

        Args:
            global_model (Model): The global model that will be federated among clients.
            strategy (Strategy): The strategy used for federated learning, specifying communication and aggregation methods.
            num_rounds (int): The number of federated learning rounds to perform.
            num_clients (int): The number of clients participating in the federated learning process.
            fed_dataset (FederatedDataset): The federated dataset used for training and evaluation.
            diff_privacy (bool, optional): Whether differential privacy is used during the federated learning process.
                                           Default is False.
        """
        self.device = torch.device(
            f"cuda" if torch.cuda.is_available() else "cpu"
        )
        self.global_model = global_model
        self.params = global_model.get_parameters()
        self.global_model.model = global_model.model.to(self.device)
        self.num_rounds = num_rounds
        self.num_clients = num_clients
        self.fed_dataset = fed_dataset
        self.strategy = strategy
        self.client_resources = client_resources
        setattr(
            self.strategy.strategy_object,
            "min_available_clients",
            self.num_clients,
        )
        setattr(
            self.strategy.strategy_object,
            "initial_parameters",
            fl.common.ndarrays_to_parameters(self.params),
        )
        setattr(self.strategy.strategy_object, "evaluate_fn", self.evaluate)
        self.fed_dataset = fed_dataset
        self.diff_priv = diff_privacy
        self.accuracies = []
        self.losses = []
        self.auc = []
        self.flower_clients = []
        self.validate()

    def validate(self) -> None:
        """Validate global_model, strategy, num_clients, num_rounds, fed_dataset, diff_privacy"""
        if not isinstance(self.global_model, Model):
            raise TypeError("global_model argument must be a Model instance")

        # if not isinstance(self.strategy, Strategy):
        #     print(self.strategy)
        #     print(isinstance(self.strategy, Strategy))
        #     raise TypeError("strategy argument must be a Strategy instance")

        if not isinstance(self.num_clients, int):
            raise TypeError("num_clients argument must be an int")

        if not isinstance(self.num_rounds, int):
            raise TypeError("num_rounds argument must be an int")

        if not isinstance(self.diff_priv, bool):
            raise TypeError("diff_priv argument must be a bool")

    def client_fn(self, cid) -> FlowerClient:
        """
        Return a FlowerClient object for a specific client ID.

        Args:
            cid: The client ID.

        Returns:
            FlowerClient: A FlowerClient object representing the individual client.
        """
        
        device = torch.device(
            f"cuda:{int(cid) % 4}" if torch.cuda.is_available() else "cpu"
        )
        client_model = copy.deepcopy(self.global_model)
      
        trainloader = self.fed_dataset.trainloaders[int(cid)]
        valloader = self.fed_dataset.valloaders[int(cid)]
        # this helps in making plots
        
        client = FlowerClient(
            cid, client_model, trainloader, valloader, self.diff_priv
        )
        self.flower_clients.append(client)
        return client

    def evaluate(
        self,
        server_round: int,
        parameters: fl.common.NDArrays,
        config: Dict[str, fl.common.Scalar],
    ) -> Optional[Tuple[float, Dict[str, fl.common.Scalar]]]:
        """
        Evaluate the global model on the validation dataset and update the accuracies and losses.

        Args:
            server_round (int): The current round of the federated learning process.
            parameters (fl.common.NDArrays): The global model parameters.
            config (Dict[str, fl.common.Scalar]): Configuration dictionary.

        Returns:
            Optional[Tuple[float, Dict[str, fl.common.Scalar]]]: The evaluation loss and accuracy.
        """
        testloader = self.fed_dataset.valloaders[0]
        
        self.global_model.set_parameters(
            parameters
        )  # Update model with the latest parameters
        loss, accuracy ,auc = self.global_model.evaluate(testloader, self.device)
        self.auc.append(auc)
        self.losses.append(loss)
        self.accuracies.append(accuracy)
        
        return loss, {"accuracy": accuracy}

    def run(self) -> None:
        """
        Run the federated learning process using Flower simulation.

        Returns:
            History: The history of the accuracies and losses during the training of each node 
        """
         # Increase the object store memory to the minimum allowed value or higher
        ray_init_args = {"include_dashboard": False
                         , "object_store_memory": 78643200
                        }
        self.fed_dataset.eng = None
        
        history = fl.simulation.start_simulation(
            client_fn=self.client_fn,
            num_clients=self.num_clients,
            config=fl.server.ServerConfig(self.num_rounds),
            strategy=self.strategy.strategy_object,
            ray_init_args=ray_init_args,
            client_resources = self.client_resources
        )

        return history




import unittest
from unittest.mock import MagicMock, patch

import pandas as pd
from sklearn.preprocessing import LabelEncoder

# Suppress pandas warnings
from scripts.base import my_eng
from MEDfl.NetManager.dataset import DataSet

from MEDfl.LearningManager.utils import params


class TestDataSet(unittest.TestCase):
    def create_mock_dataframe(
        self, path=params['path_to_test_csv']
    ):
        data_df = pd.read_csv(path)
        encoder = LabelEncoder()
        data_df["pao2fio2"].fillna(data_df["pao2fio2"].mean(), inplace=True)
        data_df["site_region"].fillna(
            data_df["site_region"].mode()[0], inplace=True
        )
        data_df["site_hospital"] = encoder.fit_transform(
            data_df["site_hospital"].astype(str)
        )
        data_df["site_region"] = encoder.fit_transform(
            data_df["site_region"].astype(str)
        )

        return data_df

    def setUp(self):
        # Create a sample DataFrame for testing

        self.sample_df = self.create_mock_dataframe()

        # Create a mock database engine
        self.mock_engine = MagicMock()

        # Initialize a DataSet object for testing
        self.dataset = DataSet(
            name="test_dataset", path=params['path_to_test_csv'], engine=self.mock_engine
        )

    def test_init(self):
        self.assertEqual(self.dataset.name, "test_dataset")
        self.assertEqual(self.dataset.path, params['path_to_test_csv'])

    def test_validate(self):
        with self.assertRaises(TypeError):
            # Invalid name type
            dataset = DataSet(name=123, path=params['path_to_test_csv'])
            dataset.validate()

        with self.assertRaises(TypeError):
            # Invalid path type
            dataset = DataSet(name="test_dataset", path=123)
            dataset.validate()

    def test_delete_dataset(self):
        # Call the delete_dataset method
        self.dataset.delete_dataset()

        # Assertions for the SQL query and execution
        self.mock_engine.execute.assert_called_once()

    def test_update_data(self):
        # update_data method is not implemented yet, so there is nothing to test
        pass

    @patch("sqlalchemy.create_engine")
    def test_list_alldatasets(self, mock_create_engine):
        # Call the list_alldatasets method
        mock_create_engine.return_value = my_eng
        self.dataset.list_alldatasets(engine=my_eng)

        # Assertions for the SQL query and execution

        self.assertTrue(mock_create_engine.called_once())


if __name__ == "__main__":
    unittest.main()



import mysql.connector
from sqlalchemy import create_engine, text
from configparser import ConfigParser
import yaml
import pkg_resources
import os

# Get the directory of the current script
current_directory = os.path.dirname(os.path.abspath(__file__))

# Load configuration from the config file
config_file_path = os.path.join(current_directory, 'db_config.ini')

config = ConfigParser()
config.read(config_file_path)
mysql_config = config['mysql']



connection_string = (
    f"mysql+mysqlconnector://{mysql_config['user']}:{mysql_config['password']}@"
    f"{mysql_config['host']}:{mysql_config['port']}/{mysql_config['database']}"
)

eng = create_engine(
    connection_string,
    execution_options={"autocommit": True},
)

my_eng = eng.connect()


from logging.config import fileConfig
import logging
from sqlalchemy import engine_from_config, create_engine
from sqlalchemy import pool
import sys
import os
from alembic import context

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from scripts.base import my_eng

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.
def configure_logger(name):
    # This is the standard logging configuration
    logging.config.fileConfig(
        'alembic_logging.ini',  # Path to your logging configuration file
        defaults={
            'logfilename': 'alembic.log',  # Log file name
        },
        disable_existing_loggers=False,
    )

    return logging.getLogger(name)



def run_migrations_offline():
    """Run migrations in 'offline' mode."""
    pass

def run_migrations_online():
    """Run migrations in 'online' mode."""
    pass

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import unittest

import pkg_resources
from sqlalchemy import text

from MEDfl.NetManager.flsetup import FLsetup
from MEDfl.NetManager.net_helper import *
from MEDfl.NetManager.net_manager_queries import *
from MEDfl.NetManager.network import Network
from MEDfl.NetManager.node import Node
from MEDfl.LearningManager.utils import params
from scripts.base import my_eng


class TestNetwork(unittest.TestCase):
    def setUp(self):
        # Create a test network
        self.network_name = "TestNetwork"
        self.network = Network(self.network_name)

    def test_create_network(self):
        # Test the create_network method
        self.network.create_network()
        # Assert that the network with the given name exists in the database
        self.assertIsNotNone(get_netid_from_name(self.network_name))
        self.network.delete_network()

    def test_delete_network(self):
        # Test the delete_network method
        # Create a test network in the database to delete (this can be done in the setUp method)
        self.network.create_network()  # Create the test network first
        self.network.delete_network()
        # Assert that the network with the given name is deleted from the database
        self.assertIsNone(get_netid_from_name(self.network_name))

    def test_update_network(self):
        # Test the update_network method
        # Create a test network in the database (this can be done in the setUp method)
        FLsetupId = 1  # Replace with a valid FLsetupId value
        self.network.create_network()  # Create the test network first
        flsetup = FLsetup(
            name="FLsetup",
            description="A random FLsetup",
            network=self.network,
        )
        flsetup.create()
        self.network.update_network(FLsetupId)
        # Assert that the FLsetupId of the network in the database is updated correctly
        updated_network = pd.read_sql(
            text(READ_NETWORK_BY_ID_QUERY),
            my_eng,
            params={"net_id": int(self.network.id)},
        ).iloc[0]

        self.assertEqual(updated_network["FLsetupId"], FLsetupId)
        self.network.delete_network()

    def test_list_allnodes(self):
        # Test the list_allnodes method
        # Create a test network in the database (this can be done in the setUp method)

        self.network.create_network()  # Create the test network first
        print(self.network.id)
        # Add some test nodes to the network using the add_node method (you can use a mock Node object)
        # For testing purposes, we assume that a Node class exists and implements the create_node method.

        mock_node = Node("test_node", 0)
        self.network.add_node(mock_node)
        # Call the list_allnodes method and check if the returned DataFrame has the expected format and data
        nodes_df = self.network.list_allnodes()
        # Assert that the DataFrame has the expected columns and data
        self.assertIsInstance(nodes_df, pd.DataFrame)

        self.assertTrue("NodeName" in nodes_df.columns.values.tolist())
        self.assertTrue("NetName" in nodes_df.columns.values.tolist())

        self.assertTrue("test_node" in nodes_df["NodeName"].iloc[0])
        self.network.delete_network()

    def test_create_master_dataset(self):
        # Test the create_master_dataset method
        # Create a test network in the database (this can be done in the setUp method)
        self.network.create_network()  # Create the test network first
        # Call the create_master_dataset method with a test CSV file (you can use a mock CSV file)
        # For testing purposes, we assume that a test CSV file 'test_dataset.csv' is available in the same directory.
        test_csv_path =  params['path_to_test_csv']
        self.network.create_master_dataset(test_csv_path)
        # Assert that the MasterDataset table is created and dataset values are inserted correctly
        self.assertTrue(master_table_exists())
        self.network.delete_network()


if __name__ == "__main__":
    unittest.main()
